#TODO List:
#Make CSS, Lua, Java, JavaScript, Batch, C, C# and C++ Language syntax highlighting.
#Allow multiple files open at the same time, with a web-browser like tab system.
#Display lists, dictionaries and tuples in debug mode.

version = "v0.98"

import string, sys, time, traceback, os
import tkinter
import tkinter.scrolledtext as tkst
import tkinter.filedialog   as tkfd
import tkinter.messagebox   as tkmb
import subprocess           as sp
from threading import Timer, Thread

root = tkinter.Tk(className="\Alex's IDE")
textpad = tkst.ScrolledText(root, width=250, height=100)

v = tkinter.StringVar()
v.set("Text")
language="Text"
currentfile=""


def getname(filepath):
    l = filepath.split("/")
    return l[len(l)-1]
def getextension(filename):
    l = filename.split(".")
    return l[1]

def _openfile():
    global currentfile
    global language
    print("Opening file...")
    filename = tkfd.askopenfilename(parent=root, title='Select a file to open.')
    if filename!="":
        with open(filename, "r",-1,"utf-8-sig") as file:
            contents=file.read()
            c = contents.replace("\r\n","\n")
            textpad.delete('1.0', 'end')
            textpad.insert('1.0', c)
            #textpad.delete('1.0','1.3') #Conversion to unicode adds 3 odd characters
            file.close()
            _highlightsyntaxs()
            print("Successfully opened the file.")
            currentfile = filename
            ext = getextension(filename)
            if ext=="py" or ext=="pyw":
                language = "Python"
                v.set("Python")
            if ext=="html" or ext=="htm" or ext=="shtml" or ext=="shtm" or ext=="xhtml" or ext=="hta":
                language = "HTML"
                v.set("HTML")
            if ext in ("bat","cmd","rd"):
                language = "Batch"
                v.set("Batch")
            root.title("Alex's IDE - "+version + " - "+getname(currentfile))
            _highlightsyntaxs()

def _save(event):
    _savefile()

def _savefile():
    global currentfile
    print("Saving file...")
    if currentfile!="":
        with open(currentfile,"w",-1,"utf-8-sig") as file:
            data = textpad.get('1.0', tkinter.END+'-1c')
            #data = data.encode("utf-8")
            file.write(data.replace("\r\n","\n"))
            file.close()
            print("Successfully saved the file.")
    else:
        _saveasfile()
        
def _saveasfile():
    global currentfile
    print("Saving file...")
    options = {}
    if(language=="Text"):
        options['defaultextension'] = '.txt'
        options['filetypes'] = [('All files', '.*'), ('Normal text file', '.txt'), ('Python Files', '.py .pyw'), ('HTML Files', '.html .htm .shtml .shtm .xhtml .hta'), ('Batch Files', '.bat .cmd .nt')]
        options['initialfile'] = 'file.txt'
    elif(language=="Python"):
        options['defaultextension'] = '.py'
        options['filetypes'] = [('Python Files', '.py .pyw')]
        options['initialfile'] = 'file.py'
    elif(language=="HTML"):
            options['defaultextension'] = '.html'
            options['filetypes'] = [('HTML Files', '.html .htm .shtml .shtm .xhtml .hta')]
            options['initialfile'] = 'file.html'
    elif(language=="Batch"):
            options['defaultextension'] = '.bat'
            options['filetypes'] = [('Batch Files', '.bat .cmd .nt')]
            options['initialfile'] = 'file.bat'

    filename = tkfd.asksaveasfilename(**options)
    if filename!="":
        with open(filename, mode="w") as file:
            data = textpad.get('1.0', tkinter.END+'-1c')
            data.replace("\r\r","\r")
            file.write(data)
            file.close()
            print("Successfully saved the file.")
            currentfile = filename
            root.title("Alex's IDE - "+version + " - "+getname(currentfile))

def _dummy():
    pass

def _run():
    if language == "Python":
        envglo = {'__name__': '__main__', '__spec__': None, '__package__': None,'__file__': currentfile}
        envloc = {}
        code = compile(textpad.get("1.0","end"), "Virtual IDE File", 'exec')
        try:
            print("")
            print("-"*60)
            print("Running file - Debug Mode")
            print("-"*60)
            print("")
            exec(code, envglo, envloc)
            print("")
            print("-"*60)
            print("Program Finished with no errors!")
            print("-"*60)
            print("Global Variables (Displaying strings, integers and floats): ")
            print("-"*60)
            for x in envglo:
                if type(envglo[x]) is int or type(envglo[x]) is str or type(envglo[x]) is float:
                    if type(envglo[x]) is str:
                        print (x,': String = "'+str(envglo[x])+'"')
                    elif type(envglo[x]) is int:
                        print (x,': Integer =',envglo[x])
                    elif type(envglo[x]) is float:
                        print (x,': Float =',envglo[x])
                    elif callable(envloc[x]):
                        print(x,': Function')
                    else:
                        print("Unknown variable.")
            print("-"*60)
            print("Local Variables (Displaying strings, integers and floats): ")
            print("-"*60)
            for x in envloc:
                if type(envloc[x]) is int or type(envloc[x]) is str or type(envloc[x]) is float:
                    if type(envloc[x]) is str:
                        print (x,': String = "'+str(envloc[x])+'"')
                    elif type(envloc[x]) is int:
                        print (x,': Integer =',envloc[x])
                    elif type(envloc[x]) is float:
                        print (x,': Float =',envloc[x])
                    elif callable(envloc[x]):
                        print(x,': Function')
                    else:
                        print("Unknown variable.")
            print("-"*60)       
        except:
            print(" ")
            print("-"*60)
            print("Error Occured!")
            print("-"*60)
            traceback.print_exc()
            print("-"*60)
            print("Global Variables (Displaying strings, integers and floats): ")
            print("-"*60)
            for x in envglo:
                if type(envglo[x]) is int or type(envglo[x]) is str or type(envglo[x]) is float:
                    if type(envglo[x]) is str:
                        print (x,': String = "'+str(envglo[x])+'"')
                    elif type(envglo[x]) is int:
                        print (x,': Integer =',envglo[x])
                    elif type(envglo[x]) is float:
                        print (x,': Float =',envglo[x])
                    elif callable(envloc[x]):
                        print(x,': Function')
                    else:
                        print("Unknown variable.")
            print("-"*60)
            print("Local Variables (Displaying strings, integers and floats): ")
            print("-"*60)
            for x in envloc:
                if type(envloc[x]) is int or type(envloc[x]) is str or type(envloc[x]) is float:
                    if type(envloc[x]) is str:
                        print (x,': String = "'+str(envloc[x])+'"')
                    elif type(envloc[x]) is int:
                        print (x,': Integer =',envloc[x])
                    elif type(envloc[x]) is float:
                        print (x,': Float =',envloc[x])
                    elif callable(envloc[x]):
                        print(x,': Function')
                    else:
                        print("Unknown variable.")
            print("-"*60)
    else:
        tkmb.showinfo(title="Message", message="Only python files can be run in debug mode, please try separate.")

def _runs():
    if language == "Python":
        envglo = {'__name__': '__main__', '__spec__': None, '__package__': None,'__file__': currentfile}
        envloc = {}
        code = compile(textpad.get("1.0","end"), "Virtual IDE File", 'exec')
        try:
            print("")
            print("-"*60)
            print("Running file - Normal Mode")
            print("-"*60)
            print("")
            exec(code, envglo, envloc)
            print("")
            print("-"*60)   
        except:
            print(" ")
            print("-"*60)
            print("Error Occured!")
            print("-"*60)
            traceback.print_exc()
            print("-"*60)
            print("Global Variables (Displaying strings, integers and floats): ")
            print("-"*60)
            for x in envglo:
                if type(envglo[x]) is int or type(envglo[x]) is str or type(envglo[x]) is float:
                    if type(envglo[x]) is str:
                        print (x,': String = "'+str(envglo[x])+'"')
                    elif type(envglo[x]) is int:
                        print (x,': Integer =',envglo[x])
                    elif type(envglo[x]) is float:
                        print (x,': Float =',envglo[x])
            print("-"*60)
            print("Local Variables (Displaying strings, integers and floats): ")
            print("-"*60)
            for x in envloc:
                if type(envloc[x]) is int or type(envloc[x]) is str or type(envloc[x]) is float:
                    if type(envloc[x]) is str:
                        print (x,': String = "'+str(envloc[x])+'"')
                    elif type(envloc[x]) is int:
                        print (x,': Integer =',envloc[x])
                    elif type(envloc[x]) is float:
                        print (x,': Float =',envloc[x])
            
            print("-"*60)
    else:
        tkmb.showinfo(title="Message", message="Only python files can be run in normal mode, please try seperate.")

def _runse():
    if language=="Python":
        if(currentfile!=""):
            _savefile()
            Thread(target=lambda: sp.call(currentfile,shell=True)).start()
        else:
            with open("temp.py", mode='w') as f:
                data = textpad.get('1.0', tkinter.END+'-1c')
                f.write(data)
                f.close()
                
                Thread(target=lambda: sp.call("temp.py",shell=True)).start()
                Timer(1,lambda: os.remove("temp.py"))
        print("-"*60)
        print("Finished file execution.")
    elif language == "HTML":
        if(currentfile!=""):
            _savefile()
            sp.call(currentfile,shell=True)
        else:
            with open("temp.html", mode='w') as f:
                data = textpad.get('1.0', tkinter.END+'-1c')
                f.write(data)
                f.close()           

                sp.call("temp.html", shell=True)
                Timer(3,lambda: os.remove("temp.html")).start()
        print("-"*60)
        print("Finished file execution.")
    elif language == "Text":
        if(currentfile!=""):
            _savefile()
            Thread(target=lambda: sp.call(currentfile,shell=True)).start()
        else:
            with open("temp.txt", mode='w') as f:
                data = textpad.get('1.0', tkinter.END+'-1c')
                f.write(data)
                f.close()           

                Thread(target=lambda: sp.call("temp.txt",shell=True)).start()
                Timer(1,lambda: os.remove("temp.txt"))
        print("-"*60)
        print("Finished file execution.")
    elif language == "Batch":
        if(currentfile!=""):
            _savefile()
            Thread(target=lambda: sp.call(currentfile,shell=False)).start()
        else:
            with open("temp.bat", mode='w') as f:
                data = textpad.get('1.0', tkinter.END+'-1c')
                f.write(data)
                f.close()           

                Thread(target=lambda: sp.call("temp.bat",shell=False)).start()
                Timer(1,lambda: os.remove("temp.bat"))
        print("-"*60)
        print("Finished file execution.")
    else:
        tkmb.showinfo(title="Message", message="Sorry, Alex IDE can't run that type of file yet.")

def _newfile():
    textpad.delete('1.0', 'end')
    global currentfile, language
    currentfile = ""
    v.set("Text")
    language="Text"
    root.title("Alex's IDE - "+version)

def _highlightsyntax(event):
    t = Timer(0.1,_highlightsyntaxs)
    t.start()

def _removehighlighting():
    textpad.tag_remove("Boolean", "1.0", tkinter.END)
    textpad.tag_remove("Statement", "1.0", tkinter.END)
    textpad.tag_remove("String", "1.0", tkinter.END)
    #textpad.tag_remove("String", "1.0", tkinter.END)
    textpad.tag_remove("Tags", "1.0", tkinter.END)
    textpad.tag_remove("InTags", "1.0", tkinter.END)
    textpad.tag_remove("Comment","1.0",tkinter.END)

def _parseindex(index):
    ys, xs = index.split(".")
    return int(ys), int(xs)

def _getindex(y,x):
    xs=str(x)
    ys=str(y)
    if x==-1:
        xs="end"
        y-=1
        ys=str(y)
    return ys+"."+xs

##################### SYNTAX HIGHLIGHTING #####################

def _highlightsyntaxs():
    _removehighlighting()
    if language == "Python":
        textpad.tag_config("Boolean", foreground="orange")
        textpad.tag_config("Statement", foreground="blue")
        textpad.tag_config("String", foreground="green")
        textpad.tag_config("Comment", foreground="purple")
        #Python Strings
        idx = '1.0'
        while 1:
            temp = idx
            idx = textpad.search("\"", temp, nocase=0, stopindex=tkinter.END)
            bidx = textpad.search("'", temp, nocase=0, stopindex=tkinter.END)
            if not idx and not bidx: break
            y = x = 1
            yb = xb = 1 
            if not idx:
                first = 1
                yb, xb = _parseindex(bidx)
            elif not bidx:
                first = 2
                y, x = _parseindex(idx)
            else:
                y, x = _parseindex(idx)
                yb, xb = _parseindex(bidx)
                lastidx = idx+'+1c'
                lastbidx =  bidx+'+1c'
                first = 0
                if y > yb:
                    first = 1
                elif y < yb:
                    first = 2
                elif x > xb:
                    first = 1
                elif x < xb:
                    first = 2
            if first == 2:
                idxa = _getindex(y, x+1)
                aidx = textpad.search("\"", idxa, nocase=0, stopindex=tkinter.END)
                if not aidx: break
                alastidx = aidx+"+1c"
                textpad.tag_add('String', idx, alastidx)
                idx = alastidx
            elif first == 1:
                idxa = _getindex(yb, xb+1)
                aidx = textpad.search("'", idxa, nocase=0, stopindex=tkinter.END)
                if not aidx: break
                alastidx = aidx+"+1c"
                textpad.tag_add('String', bidx, alastidx)
                idx = alastidx
            aidx = alastidx
        #Python Commenting
        idx = '1.0'
        while 1:
            idx = textpad.search("#", idx, nocase=0, stopindex=tkinter.END)
            if not idx: break
            y,x=_parseindex(idx)
            lastidx = str(y)+".end"
            textpad.tag_add("Comment",idx,lastidx)
            idx = lastidx
        #Python Booleans
        for s in ("True", "False", "None"):
            idx = '1.0'
            while 1:
                idx = textpad.search(s, idx, nocase=0, stopindex=tkinter.END)
                if not idx: break
                lastidx = '%s+%dc' % (idx, len(s))
                
                cont1 = False
                cont2 = False
                for l in (" ","","(","=",string.whitespace[1],string.whitespace[2]):
                    y, x = _parseindex(idx)
                    idxa = _getindex(y, x-1)
                    bletter = textpad.get(idxa,idx)
                    if bletter==l:
                        cont1=True
                        break
                for l in (" ","",")",";"):
                    y, x = _parseindex(idx)
                    lastidxa = _getindex(y, x+1+len(l))
                    aletter = textpad.get(lastidx,lastidxa)
                    if aletter==l:
                        cont2=True
                        break
                if cont1 and cont2:
                    textpad.tag_add('Boolean', idx, lastidx)
                idx = lastidx
        #Python Keywords
        for s in ("input","print","int","float","str","if", "for","while", "try", "except", "elif", "else", "break", "continue", "import", "from", "as", "with", "in", "raise", "return", "assert", "nonlocal", "def", "pass", "class", "finally", "lambda", "del","global", "is","and","or","not","yield"):
            idx = '1.0'
            while 1:
                idx = textpad.search(s, idx, nocase=0, stopindex=tkinter.END)
                if not idx: break
                y,x = _parseindex(idx)
                lastidx = _getindex(y,x+len(s))
                cont1 = False
                cont2 = False
                for l in (" ","","(",string.whitespace[1],string.whitespace[2]):
                    y, x = _parseindex(idx)
                    idxa = _getindex(y, x-1)
                    bletter = textpad.get(idxa,idx)
                    if bletter==l:
                        cont1=True
                        break
                for l in (" ","","(",":",";", string.whitespace[1],string.whitespace[2]):
                    y, x = _parseindex(lastidx)
                    lastidxa = _getindex(y, x+1)
                    aletter = textpad.get(lastidx,lastidxa) 
                    if aletter==l:
                        cont2=True
                        break
                y, x = _parseindex(idx)
                if cont1 and cont2:
                    textpad.tag_add('Statement', idx, lastidx)
                idx = lastidx
    elif language == "HTML":
        textpad.tag_config("Tags", foreground="red")
        textpad.tag_config("InTags", foreground="blue")
        textpad.tag_config("String", foreground="green")
        idx = '1.0'
        while 1: 
            idx = textpad.search("<", idx, nocase=0, stopindex=tkinter.END)
            if not idx: break
            lastidx = '%s+%dc' % (idx, 1)
            y, x = _parseindex(idx)
            tag1a = _getindex(y, x)
            tag1b = _getindex(y, x+1)
            idxa = _getindex(y, x+1)
            intaga = _getindex(y, x+1)
            aidx = textpad.search(">", idxa, nocase=0, stopindex=tkinter.END)
            if not aidx: break
            y, x = _parseindex(aidx)
            tag2a = _getindex(y, x)
            tag2b = _getindex(y, x+1)
            alastidx = '%s+%dc' % (aidx, 1)
            intagb = _getindex(y, x)
            textpad.tag_add('Tags', tag1a, tag1b)
            textpad.tag_add('Tags', tag2a, tag2b)
            textpad.tag_add('InTags', intaga, intagb)
            idx = alastidx
            aidx = alastidx
    elif language == "Batch":
        textpad.tag_config("Statement",foreground="blue")
        textpad.tag_config("String",foreground="green")
        textpad.tag_config("Comment",foreground="purple")
        #Batch strings
        idx = '1.0'
        while 1:
            temp = idx
            idx = textpad.search("\"", temp, nocase=0, stopindex=tkinter.END)
            bidx = textpad.search("'", temp, nocase=0, stopindex=tkinter.END)
            if not idx and not bidx: break
            y = x = 1
            yb = xb = 1 
            if not idx:
                first = 1
                yb, xb = _parseindex(bidx)
            elif not bidx:
                first = 2
                y, x = _parseindex(idx)
            else:
                y, x = _parseindex(idx)
                yb, xb = _parseindex(bidx)
                lastidx = idx+'+1c'
                lastbidx =  bidx+'+1c'
                first = 0
                if y > yb:
                    first = 1
                elif y < yb:
                    first = 2
                elif x > xb:
                    first = 1
                elif x < xb:
                    first = 2
            if first == 2:
                idxa = _getindex(y, x+1)
                aidx = textpad.search("\"", idxa, nocase=0, stopindex=tkinter.END)
                if not aidx: break
                alastidx = aidx+"+1c"
                textpad.tag_add('String', idx, alastidx)
                idx = alastidx
            elif first == 1:
                idxa = _getindex(yb, xb+1)
                aidx = textpad.search("'", idxa, nocase=0, stopindex=tkinter.END)
                if not aidx: break
                alastidx = aidx+"+1c"
                textpad.tag_add('String', bidx, alastidx)
                idx = alastidx
            aidx = alastidx
        #Batch commenting
        idx = '1.0'
        while 1:
            idx = textpad.search("rem", idx, nocase=0, stopindex=tkinter.END)
            if not idx: break
            y,x=_parseindex(idx)
            lastidx = str(y)+".end"
            textpad.tag_add("Comment",idx,lastidx)
            idx = lastidx
        #Batch statements
        for s in ("tracert","ping","assoc","attrib","break","bcdedit","cacls","call","cd","chcp","chdir","chkdsk","chkntfs","cls","cmd","color","comp","compact","convert","copy","date","del","dir","diskpart","doskey","driverquery","echo","endlocal","erase","exit","fc","find","findstr","for","format","fsutil","ftype","goto","gpresult","graftabl","help","icacls","if","label","md","mkdir","mklink","mode","more","move","openfiles","path","pause","popd","print","prompt","pushd","rd","recover","ren","rename","replace","rmdir","robocopy","set","setlocal","sc","schtasks","shift","shutdown","sort","start","subst","systeminfo","tasklist","taskkill","time","title","tree","type","ver","verify","vol","xcopy","wmic"):
            idx = '1.0'
            while 1:
                idx = textpad.search(s, idx, nocase=0, stopindex=tkinter.END)
                if not idx: break
                y,x = _parseindex(idx)
                lastidx = _getindex(y,x+len(s))
                cont1 = False
                cont2 = False
                for l in (" ","","@",string.whitespace[1],string.whitespace[2]):
                    y, x = _parseindex(idx)
                    if(x==0):
                        cont1=True
                        break
                    idxa = _getindex(y, x-1)
                    bletter = textpad.get(idxa,idx)
                    if bletter==l:
                        cont1=True
                        break
                for l in (" ","", string.whitespace[1],string.whitespace[2]):
                    y, x = _parseindex(lastidx)
                    lastidxa = _getindex(y, x+1)
                    aletter = textpad.get(lastidx,lastidxa) 
                    if aletter==l:
                        cont2=True
                        break
                if cont1 and cont2:
                    textpad.tag_add('Statement', idx, lastidx)
                idx = lastidx
##################### SYNTAX HIGHLIGHTING END #####################
def _tab(e):
    textpad.insert(tkinter.INSERT," "*4)
    return 'break'

def _newlang():
    global language
    global v
    language = v.get()
    _highlightsyntaxs()

def _exit():
    try:
        if tkmb.askokcancel("Quit", "Are you sure you want to quit?"):
            root.destroy()
    except:
        sys.exit(1)

print("Welcome to Alex IDE ("+version+").")
print("This is the console window.")
print("")
print("-"*60)
menu = tkinter.Menu(root)
root.config(menu=menu)
filemenu = tkinter.Menu(menu,tearoff=0)
runmenu = tkinter.Menu(menu)
langmenu = tkinter.Menu(menu,tearoff=0)

menu.add_cascade(label="File",menu=filemenu)

filemenu.add_command(label="New", command=_newfile)
filemenu.add_command(label="Open...", command=_openfile)
filemenu.add_command(label="Save", command=_savefile)
filemenu.add_command(label="Save As...", command=_saveasfile)
filemenu.add_separator()
filemenu.add_command(label="Exit", command=_exit)

#menu.add_cascade(label="Help", menu=helpmenu)
#helpmenu.add_command(label="Test Function", command=_highlightsyntaxs)


menu.add_cascade(label="Language", menu=langmenu)

langmenu.add_radiobutton(label="Plain Text", command=_newlang, variable=v, value="Text")
langmenu.add_radiobutton(label="Python", command=_newlang, variable=v, value="Python")
langmenu.add_radiobutton(label="HTML", command=_newlang, variable=v, value="HTML")
langmenu.add_radiobutton(label="Batch", command=_newlang,variable=v,value="Batch")
langmenu.add_separator()
langmenu.add_command(label="More coming soon.", command=_dummy)

menu.add_cascade(label="Run", menu=runmenu)
runmenu.add_command(label="Run - Debug", command=_run)
runmenu.add_command(label="Run - Normal", command=_runs)
runmenu.add_command(label="Run - Separate", command=_runse)

textpad.bind("<Key>", _highlightsyntax)
textpad.bind("<Control-s>", _save)
textpad.bind("<Tab>",_tab)
textpad.pack()
root.geometry("800x600+400+150")
root.title("Alex's IDE - "+version)
root.protocol("WM_DELETE_WINDOW", _exit)
root.iconbitmap("icon.ico")
root.mainloop()
